from cone.app.security import PrincipalACL
from node.ext.zodb import OOBTNodeAttributes
from node.utils import instance_property
from plumber import default
from plumber import plumb


class ZODBPrincipalACL(PrincipalACL):
    """Principal ACL for ZODB nodes.
    """

    @plumb
    def __init__(next_, self, *a, **kw):
        next_(self, *a, **kw)
        self.principal_roles

    @default
    @instance_property
    def principal_roles(self):
        return OOBTNodeAttributes(name='principal_roles')


class ZODBEntryPrincipalACL(PrincipalACL):

    @default
    @property
    def principal_roles(self):
        return self.storage.principal_roles



from cone.app.model import BaseNode
from cone.app.model import Metadata
from cone.app.model import Properties
from cone.zodb import testing
from cone.zodb import zodb_entry_for
from cone.zodb import ZODBEntry
from cone.zodb import ZODBEntryNode
from cone.zodb.interfaces import IZODBEntry
from cone.zodb.interfaces import IZODBEntryNode
from cone.zodb.testing import ZODBDummyNode
from node.ext.zodb import OOBTNodeAttributes
from node.interfaces import IOrder
from node.interfaces import IOrdered
from node.interfaces import IUUIDAware
from node.tests import NodeTestCase
from node.utils import LocationIterator
import transaction
import uuid


class TestEntry(NodeTestCase):
    layer = testing.zodb_layer

    def test_entry(self):
        self.layer.new_request()

        # Create node containing ZODBEntry
        root = BaseNode(name='root')
        root['myentry'] = ZODBEntry()
        entry = root['myentry']
        self.assertTrue(isinstance(entry, ZODBEntry))

        # ZODB entry node of entry is looked up by db_name from db root
        self.assertTrue(isinstance(entry.storage, ZODBEntryNode))
        self.assertEqual(entry.storage.name, 'myentry')

        # storage, entry and parents
        self.assertTrue(entry.storage._v_parent is entry)
        self.assertTrue(entry.storage.entry is entry)

        self.assertTrue(entry.storage._v_parent.parent is root)
        self.assertTrue(entry.storage.__parent__ is root)
        self.assertTrue(entry.storage.parent is root)

        # ``metadata`` and ``properties`` are returned from entry
        self.assertTrue(isinstance(entry.storage.metadata, Metadata))
        self.assertTrue(isinstance(entry.storage.properties, Properties))

        self.assertTrue(entry.storage.metadata is entry.metadata)
        self.assertTrue(entry.storage.properties is entry.properties)

        # ``attrs`` are returned from storage
        self.assertTrue(isinstance(entry.storage.attrs, OOBTNodeAttributes))
        self.assertTrue(entry.attrs is entry.storage.attrs)
        self.assertTrue(entry.attrs.parent is entry.storage)

        # Create children
        foo = ZODBDummyNode()
        entry['foo'] = foo
        bar = ZODBDummyNode()
        bar.attrs['title'] = 'bar'
        entry['bar'] = bar

        # ``__iter__``
        self.assertEqual([k for k in entry], ['foo', 'bar'])

        # ``__getitem__``
        self.assertTrue(entry['foo'] is foo)

        # ``keys``
        self.assertEqual(foo.attrs.keys(), ['title', 'uuid'])

        # ``__parent__``
        self.assertTrue(foo.parent is entry)
        self.assertTrue(foo.parent.parent is root)

        # IZODBEntry and IZODBEntryNode
        self.assertTrue(IZODBEntry.providedBy(entry))
        self.assertTrue(IZODBEntryNode.providedBy(entry.storage))

        # ZODBDummyNode is UUIDAware
        self.assertTrue(IUUIDAware.providedBy(foo))
        self.assertTrue(isinstance(foo.uuid, uuid.UUID))

        # Entry and entry node result in the same tree
        self.checkOutput("""
        <class 'cone.zodb.entry.ZODBEntry'>: myentry
          <class 'cone.zodb.testing.ZODBDummyNode'>: foo
          <class 'cone.zodb.testing.ZODBDummyNode'>: bar
        """, entry.treerepr())

        self.checkOutput("""
        <class 'cone.zodb.entry.ZODBEntryNode'>: myentry
          <class 'cone.zodb.testing.ZODBDummyNode'>: foo
          <class 'cone.zodb.testing.ZODBDummyNode'>: bar
        """, entry.storage.treerepr())

        # IOrdered and IOrder
        self.assertTrue(IOrdered.providedBy(entry))
        self.assertTrue(IOrder.providedBy(entry))

        # first_key
        self.assertEqual(entry.first_key, 'foo')

        # last_key
        self.assertEqual(entry.last_key, 'bar')

        # next_key
        self.assertEqual(entry.next_key('foo'), 'bar')

        # prev_key
        self.assertEqual(entry.prev_key('bar'), 'foo')

        # swap
        entry.swap(entry['foo'], entry['bar'])
        self.assertEqual(entry.keys(), ['bar', 'foo'])

        # insertbefore
        foo = entry.detach('foo')
        entry.insertbefore(foo, entry['bar'])
        self.assertEqual(entry.keys(), ['foo', 'bar'])

        # insertafter
        foo = entry.detach('foo')
        entry.insertafter(foo, entry['bar'])
        self.assertEqual(entry.keys(), ['bar', 'foo'])

        # insertfirst
        foo = entry.detach('foo')
        entry.insertfirst(foo)
        self.assertEqual(entry.keys(), ['foo', 'bar'])

        # insertlast
        foo = entry.detach('foo')
        entry.insertlast(foo)
        self.assertEqual(entry.keys(), ['bar', 'foo'])

        # movebefore
        entry.movebefore('foo', 'bar')
        self.assertEqual(entry.keys(), ['foo', 'bar'])

        # moveafter
        entry.moveafter('foo', 'bar')
        self.assertEqual(entry.keys(), ['bar', 'foo'])

        # movefirst
        entry.movefirst('foo')
        self.assertEqual(entry.keys(), ['foo', 'bar'])

        # movelast
        entry.movelast('foo')
        self.assertEqual(entry.keys(), ['bar', 'foo'])

        # ``__delitem__``
        del entry['foo']
        self.checkOutput("""
        <class 'cone.zodb.entry.ZODBEntry'>: myentry
          <class 'cone.zodb.testing.ZODBDummyNode'>: bar
        """, entry.treerepr())

        # ``__call__`` delegates to storage, which is the ZODB entry node
        entry()

        # ``zodb_entry_for``
        self.assertTrue(zodb_entry_for(entry) is entry)
        self.assertTrue(zodb_entry_for(entry.storage) is entry)
        self.assertTrue(zodb_entry_for(bar) is entry)
        self.assertTrue(zodb_entry_for(root) is None)

        # LocationIterator
        self.assertEqual([_ for _ in LocationIterator(bar)], [bar, entry, root])
        self.assertEqual([_ for _ in LocationIterator(entry)], [entry, root])

        # DB name
        class CustomZODBEntry(ZODBEntry):
            @property
            def db_name(self):
                return 'custom_entry_storage'

            @property
            def name(self):
                return 'entry_storage'

        root['custom_entry_storage'] = CustomZODBEntry(name='custom_entry')
        entry = root['custom_entry_storage']
        self.assertEqual(entry.name, 'entry_storage')

        child = ZODBDummyNode()
        entry['child'] = child
        child = entry['child']
        self.assertEqual(child.path, ['root', 'entry_storage', 'child'])
        self.assertEqual(entry.db_name, 'custom_entry_storage')

        transaction.commit()



from BTrees.IFBTree import IFSet
from cone.app.model import BaseNode
from cone.zodb import CatalogProxy
from cone.zodb import testing
from cone.zodb.catalog import CatalogIndexer
from cone.zodb.indexing import create_default_catalog
from cone.zodb.indexing import create_default_metadata
from cone.zodb.testing import CatalogAwareZODBEntry
from cone.zodb.testing import CatalogAwareZODBNode
from cone.zodb.testing import CatalogIncludedZODBEntry
from node.tests import NodeTestCase
from persistent import Persistent
from repoze.catalog.catalog import Catalog
from repoze.catalog.indexes.field import CatalogFieldIndex
from repoze.catalog.indexes.path import CatalogPathIndex
from repoze.catalog.query import Eq
import transaction


class TestCatalog(NodeTestCase):
    layer = testing.zodb_layer

    def test_catalog(self):
        # Create calatog aware ZODB entry
        self.layer.new_request()

        root = BaseNode(name='root')
        entry = root['catalog_aware'] = CatalogAwareZODBEntry()
        catalog = entry.catalog_proxies['default'].catalog

        self.assertTrue(isinstance(catalog, Catalog))
        self.assertEqual(
            sorted(catalog.keys()),
            ['app_path', 'path', 'state', 'title', 'type', 'uid']
        )
        self.assertTrue(isinstance(catalog['app_path'], CatalogPathIndex))
        self.assertTrue(isinstance(catalog['path'], CatalogPathIndex))
        self.assertTrue(isinstance(catalog['state'], CatalogFieldIndex))
        self.assertTrue(isinstance(catalog['title'], CatalogFieldIndex))
        self.assertTrue(isinstance(catalog['type'], CatalogFieldIndex))
        self.assertTrue(isinstance(catalog['uid'], CatalogFieldIndex))

        # Empty node. not added yet
        foo = CatalogAwareZODBNode()
        uid = foo.uuid
        res = entry.catalog_proxies['default'].catalog.query(Eq('uid', uid))
        self.assertEqual(res[0], 0)
        self.assertTrue(isinstance(res[1], IFSet))
        self.assertEqual(len(res[1]), 0)

        # Add nodes and query catalog
        bar = CatalogAwareZODBNode()
        bar.attrs['title'] = 'bar'

        entry['foo'] = foo
        entry['bar'] = bar

        res = entry.catalog_proxies['default'].catalog.query(Eq('uid', uid))
        self.assertEqual(res[0], 1)
        self.assertEqual(len(res[1]), 1)

        # Check path index
        res = entry.catalog_proxies['default'].catalog.query(
            Eq('path', {'query': '/catalog_aware', 'level': 0})
        )
        self.assertEqual(res[0], 2)
        self.assertEqual(len(res[1]), 2)

        res = entry.catalog_proxies['default'].catalog.query(
            Eq('app_path', {'query': 'root'})
        )
        self.assertEqual(res[0], 2)
        self.assertEqual(len(res[1]), 2)

        # Check metadata
        self.assertEqual(
            sorted(entry.catalog_proxies['default'].doc_metadata(uid).items()),
            [
                ('app_path', ['root', 'catalog_aware', 'foo']),
                ('combined_title', 'catalog_aware - foo'),
                ('path', ['catalog_aware', 'foo']),
                ('title', 'foo'),
                ('uid', uid)]
        )

        # Reindexing happens at ``__call__`` time
        foo.attrs['title'] = 'foo changed'
        foo()
        self.assertEqual(
            sorted(entry.catalog_proxies['default'].doc_metadata(str(uid)).items()),
            [
                ('app_path', ['root', 'catalog_aware', 'foo']),
                ('combined_title', 'catalog_aware - foo changed'),
                ('path', ['catalog_aware', 'foo']),
                ('title', 'foo changed'),
                ('uid', uid)
            ]
        )

        # Calling the ZODB entry delegates to refering ZODB entry node
        entry()

        # Create child for 'bar'
        child = CatalogAwareZODBNode()
        bar['child'] = child
        child.attrs['title'] = 'Child of bar'
        child()

        self.checkOutput("""
        <class 'cone.zodb.testing.CatalogAwareZODBEntry'>: catalog_aware
          <class 'cone.zodb.testing.CatalogAwareZODBNode'>: foo
          <class 'cone.zodb.testing.CatalogAwareZODBNode'>: bar
            <class 'cone.zodb.testing.CatalogAwareZODBNode'>: child
        """, entry.treerepr())

        bar_uid = bar.uuid
        res = entry.catalog_proxies['default'].catalog.query(Eq('uid', bar_uid))
        self.assertEqual(res[0], 1)
        self.assertEqual(len(res[1]), 1)

        child_uid = child.uuid
        res = entry.catalog_proxies['default'].catalog.query(Eq('uid', child_uid))
        self.assertEqual(res[0], 1)
        self.assertEqual(len(res[1]), 1)

        # Rebuild catalog
        self.assertEqual(
            entry.catalog_proxies['default'].rebuild_catalog(),
            3
        )

        res = entry.catalog_proxies['default'].catalog.query(Eq('type', 'dummytype'))
        self.assertEqual(res[0], 3)
        self.assertEqual(len(res[1]), 3)

        # Delete node. Gets unindexed recursive.
        del entry['bar']

        self.checkOutput("""
        <class 'cone.zodb.testing.CatalogAwareZODBEntry'>: catalog_aware
          <class 'cone.zodb.testing.CatalogAwareZODBNode'>: foo
        """, entry.treerepr())

        res = entry.catalog_proxies['default'].catalog.query(Eq('uid', bar_uid))
        self.assertEqual(res[0], 0)
        self.assertEqual(len(res[1]), 0)

        res = entry.catalog_proxies['default'].catalog.query(Eq('uid', child_uid))
        self.assertEqual(res[0], 0)
        self.assertEqual(len(res[1]), 0)

        res = entry.catalog_proxies['default'].catalog.query(Eq('type', 'dummytype'))
        self.assertEqual(res[0], 1)
        self.assertEqual(len(res[1]), 1)

        self.assertEqual(
            entry.catalog_proxies['default'].rebuild_catalog(),
            1
        )

        # Test moving of subtrees, if objects get indexed the right way
        source = entry['source'] = CatalogAwareZODBNode()
        source['c1'] = CatalogAwareZODBNode()
        source['c2'] = CatalogAwareZODBNode()
        entry['target'] = CatalogAwareZODBNode()
        self.checkOutput("""
        <class 'cone.zodb.testing.CatalogAwareZODBEntry'>: catalog_aware
          <class 'cone.zodb.testing.CatalogAwareZODBNode'>: foo
          <class 'cone.zodb.testing.CatalogAwareZODBNode'>: source
            <class 'cone.zodb.testing.CatalogAwareZODBNode'>: c1
            <class 'cone.zodb.testing.CatalogAwareZODBNode'>: c2
          <class 'cone.zodb.testing.CatalogAwareZODBNode'>: target
        """, entry.treerepr())

        uid = source['c1'].uuid
        self.assertEqual(
            sorted(entry.catalog_proxies['default'].doc_metadata(str(uid)).items()),
            [
                ('app_path', ['root', 'catalog_aware', 'source', 'c1']),
                ('combined_title', 'catalog_aware - foo - foo'),
                ('path', ['catalog_aware', 'source', 'c1']),
                ('title', 'foo'),
                ('uid', uid)
            ]
        )

        to_move = entry.detach('source')
        target = entry['target']
        target[to_move.name] = to_move
        uid = target['source']['c1'].uuid
        self.assertEqual(
            sorted(entry.catalog_proxies['default'].doc_metadata(str(uid)).items()),
            [
                ('app_path', ['root', 'catalog_aware', 'target', 'source', 'c1']),
                ('combined_title', 'catalog_aware - foo - foo - foo'),
                ('path', ['catalog_aware', 'target', 'source', 'c1']),
                ('title', 'foo'),
                ('uid', uid)
            ]
        )

        self.checkOutput("""
        <class 'cone.zodb.testing.CatalogAwareZODBEntry'>: catalog_aware
          <class 'cone.zodb.testing.CatalogAwareZODBNode'>: foo
          <class 'cone.zodb.testing.CatalogAwareZODBNode'>: target
            <class 'cone.zodb.testing.CatalogAwareZODBNode'>: source
              <class 'cone.zodb.testing.CatalogAwareZODBNode'>: c1
              <class 'cone.zodb.testing.CatalogAwareZODBNode'>: c2
        """, entry.treerepr())

        res = entry.catalog_proxies['default'].catalog.query(
            Eq('path', {'query': '/catalog_aware/target'})
        )
        self.assertEqual(res[0], 4)
        self.assertEqual(len(res[1]), 4)

        res = entry.catalog_proxies['default'].catalog.query(
            Eq('path', {'query': '/catalog_aware/target/source'})
        )
        self.assertEqual(res[0], 3)
        self.assertEqual(len(res[1]), 3)

        # Test CatalogIndexer
        indexer = entry.catalog_indexer
        self.assertTrue(isinstance(indexer, CatalogIndexer))

        foo = entry['foo']
        indexer.unindex_doc(foo)

        res = entry.catalog_proxies['default'].catalog.query(Eq('uid', foo.uuid))
        self.assertEqual(res[0], 0)
        self.assertEqual(len(res[1]), 0)

        indexer.index_doc(foo)
        res = entry.catalog_proxies['default'].catalog.query(Eq('uid', foo.uuid))
        self.assertEqual(res[0], 1)
        self.assertEqual(len(res[1]), 1)

        target = entry['target']
        indexer.unindex_recursiv(target)
        res = entry.catalog_proxies['default'].catalog.query(
            Eq('path', {'query': '/catalog_aware'})
        )
        self.assertEqual(res[0], 1)
        self.assertEqual(len(res[1]), 1)

        indexer.index_recursiv(target)
        res = entry.catalog_proxies['default'].catalog.query(
            Eq('path', {'query': '/catalog_aware'})
        )
        self.assertEqual(res[0], 5)
        self.assertEqual(len(res[1]), 5)

        # Test pointing CatalogProxy to existing objects providing the wrong
        # interface
        db_root = self.layer.zodb_root()
        db_root['invalid_catalog_object'] = Persistent()
        db_root['invalid_doc_map_object'] = Persistent()

        proxy = CatalogProxy(
            entry['foo'],
            entry,
            'invalid_catalog_object',
            'invalid_doc_map_object',
            create_default_catalog,
            create_default_metadata
        )

        err = self.expectError(
            ValueError,
            lambda: proxy.catalog
        )
        expected = 'ICatalog not provided by invalid_catalog_object'
        self.assertEqual(str(err), expected)

        err = self.expectError(
            ValueError,
            lambda: proxy.doc_map
        )
        expected = 'invalid_catalog_object not a DocumentMap instance'
        self.assertEqual(str(err), expected)

        entry = root['catalog_included'] = CatalogIncludedZODBEntry()
        entry.attrs['title'] = 'Calatog Included'
        uid = entry.storage.uuid
        res = entry.catalog_proxies['default'].catalog.query(Eq('uid', uid))
        self.assertEqual(res[0], 1)
        self.assertEqual(len(res[1]), 1)

        self.assertEqual([
            ('app_path', ['root', 'catalog_included']),
            ('combined_title', 'catalog_included'),
            ('path', ['catalog_included']),
            ('title', 'catalog_included'),
            ('uid', uid)
        ], list(entry.catalog_proxies['default'].doc_metadata(str(uid)).items()))

        # Re-init ZODB connection
        transaction.commit()
        self.layer.init_zodb()



from cone.app.model import AppNode
from cone.app.model import UUIDAttributeAware
from cone.app.security import DEFAULT_ACL
from cone.app.testing import Security
from cone.zodb import CatalogAware
from cone.zodb import CatalogProvidingEntry
from cone.zodb import ZODBEntry
from cone.zodb import ZODBEntryNode
from cone.zodb import ZODBEntryPrincipalACL
from cone.zodb import ZODBPrincipalACL
from node.ext.zodb import OOBTNode
from node.ext.zodb import ZODBNode
from plumber import plumbing
from ZODB.blob import BlobStorage
from ZODB.DB import DB
from ZODB.FileStorage import FileStorage
import os
import shutil
import tempfile


@plumbing(UUIDAttributeAware)
class ZODBDummyNode(OOBTNode):
    node_info_name = 'dummytype'

    def __init__(self, name=None, parent=None):
        OOBTNode.__init__(self, name=name, parent=parent)
        self.attrs['title'] = 'foo'


@plumbing(CatalogAware)
class CatalogAwareZODBNode(ZODBDummyNode):
    pass


@plumbing(CatalogAware)
class CatalogAwareZODBEntryNode(ZODBEntryNode):
    pass


@plumbing(CatalogProvidingEntry)
class CatalogAwareZODBEntry(ZODBEntry):
    node_factory = CatalogAwareZODBEntryNode


@plumbing(CatalogAware, UUIDAttributeAware)
class CatalogIncludedZODBEntryNode(ZODBEntryNode):
    include_entry = True


@plumbing(CatalogProvidingEntry)
class CatalogIncludedZODBEntry(ZODBEntry):
    node_factory = CatalogIncludedZODBEntryNode


@plumbing(AppNode, ZODBPrincipalACL)
class ZODBPrincipalACLNode(ZODBNode):
    pass


@plumbing(ZODBPrincipalACL)
class ZODBPrincipalACLEntryNode(ZODBEntryNode):

    @property
    def __acl__(self):
        return DEFAULT_ACL


@plumbing(ZODBEntryPrincipalACL)
class ZODBPrincipalACLEntry(ZODBEntry):
    node_factory = ZODBPrincipalACLEntryNode


class ZODBLayer(Security):

    def setUp(self, args=None):
        self.tempdir = tempfile.mkdtemp()
        self.zodb_connection = None
        self.init_zodb()
        super(ZODBLayer, self).setUp(args=args)

    def tearDown(self):
        super(ZODBLayer, self).tearDown()
        self.zodb.close()
        shutil.rmtree(self.tempdir)

    def new_request(self, type=None, xhr=False):
        request = super(ZODBLayer, self).new_request(type=type, xhr=xhr)
        setattr(request, '_primary_zodb_conn', self.zodb_connection)
        return request

    def init_zodb(self):
        if hasattr(self, 'zodb') and self.zodb:
            self.zodb_connection.close()
            self.zodb.close()
        filestorage_dir = os.path.join(self.tempdir, 'Data.fs')
        filestorage = FileStorage(filestorage_dir)
        blobstorage_dir = os.path.join(self.tempdir, 'blobstorage')
        blobstorage = BlobStorage(
            blobstorage_dir,
            filestorage,
            layout='automatic'
        )
        self.zodb = DB(blobstorage)
        self.zodb_connection = self.zodb.open()

    def zodb_root(self):
        return self.zodb_connection.root()


zodb_layer = ZODBLayer()



from cone.app.model import AppNode
from cone.zodb.interfaces import IZODBEntry
from cone.zodb.interfaces import IZODBEntryNode
from node.behaviors import DefaultInit
from node.behaviors import Lifecycle
from node.behaviors import MappingConstraints
from node.behaviors import MappingNode
from node.behaviors import MappingStorage
from node.behaviors import Order
from node.ext.zodb import IZODBNode
from node.ext.zodb import OOBTNode
from node.interfaces import IOrdered
from node.locking import locktree
from plumber import default
from plumber import override
from plumber import plumbing
from pyramid.threadlocal import get_current_request
from pyramid_zodbconn import get_connection
from zope.interface import implementer


def zodb_entry_for(node):
    while node:
        if IZODBEntry.providedBy(node):
            return node
        if IZODBEntryNode.providedBy(node):
            return node.entry
        if not IZODBNode.providedBy(node) or node.parent is None:
            return None
        node = node.parent


@implementer(IZODBEntryNode)
class ZODBEntryNode(OOBTNode):

    @property
    def __parent__(self):
        # ZODBEntryStorage instance is set as parent, but we are
        # interested in it's parent when traversing.
        return self._v_parent.parent

    @property
    def entry(self):
        return self._v_parent

    @property
    def metadata(self):
        return self.entry.metadata

    @property
    def properties(self):
        return self.entry.properties

    def __getitem__(self, name):
        v = super(ZODBEntryNode, self).__getitem__(name)
        # _v_parent is not set if ZODBEntryNode is not read via ZODBEntryStorage
        # but from ZODB root directly.
        v._v_parent = getattr(self, '_v_parent', None)
        return v


@implementer(IZODBEntry)
class ZODBEntryStorage(MappingStorage):
    node_factory = default(ZODBEntryNode)

    @default
    @property
    def db_name(self):
        return self.name

    @default
    @property
    def db_root(self):
        # XXX: currently primary DB only, support named DB
        #      currently requires a current request. make connection lookup
        #      function searching on request first, but provides fallback.
        conn = get_connection(get_current_request())
        return conn.root()

    @default
    @property
    def storage(self):
        storage = self.db_root.get(self.db_name)
        if not storage:
            storage = self.node_factory(name=self.name, parent=self)
            self.db_root[self.db_name] = storage
        else:
            storage._v_parent = self
        return storage

    @override
    @property
    def attrs(self):
        return self.storage.attrs

    @override
    @locktree
    def __setitem__(self, key, val):
        self.storage[key] = val

    @override
    @locktree
    def __delitem__(self, key):
        del self.storage[key]

    @default
    @locktree
    def __call__(self):
        self.storage()


@plumbing(
    AppNode,
    MappingConstraints,
    DefaultInit,
    Order,
    MappingNode,
    Lifecycle,
    ZODBEntryStorage)
@implementer(IOrdered)
class ZODBEntry(object):

    # override methods of ``Order`` which expect odict like storage on self.
    # we already have ``IOrder`` providing node as storage, proxy related
    # calls to storage.

    def swap(self, node_a, node_b):
        self.storage.swap(node_a, node_b)

    def insertbefore(self, newnode, refnode):
        self.storage.insertbefore(newnode, refnode)

    def insertafter(self, newnode, refnode):
        self.storage.insertafter(newnode, refnode)

    def insertfirst(self, newnode):
        self.storage.insertfirst(newnode)

    def insertlast(self, newnode):
        self.storage.insertlast(newnode)

    def movebefore(self, movenode, refnode):
        self.storage.movebefore(movenode, refnode)

    def moveafter(self, movenode, refnode):
        self.storage.moveafter(movenode, refnode)

    def movefirst(self, movenode):
        self.storage.movefirst(movenode)

    def movelast(self, movenode):
        self.storage.movelast(movenode)



from cone.zodb import testing
from cone.zodb.testing import ZODBPrincipalACLEntry
from cone.zodb.testing import ZODBPrincipalACLNode
from cone.zodb.testing import ZODBPrincipalACLEntryNode
from node.ext.zodb import OOBTNodeAttributes
from node.tests import NodeTestCase
from pyramid.security import ALL_PERMISSIONS


class TestCommon(NodeTestCase):
    layer = testing.zodb_layer

    def test_ZODBPrincipalACL(self):
        node = ZODBPrincipalACLNode()
        self.assertTrue(isinstance(node.principal_roles, OOBTNodeAttributes))
        self.assertEqual(node.principal_roles.name, 'principal_roles')

        node.principal_roles['someuser'] = ['manager']
        acl = node.__acl__

        self.assertEqual(acl[0][0], 'Allow')
        self.assertEqual(acl[0][1], 'someuser')
        self.assertEqual(sorted(acl[0][2]), [
            'add', 'change_order', 'change_state', 'copy', 'cut', 'delete',
            'edit', 'list', 'manage', 'manage_permissions', 'paste', 'view'
        ])

        self.assertEqual(acl[1][0], 'Allow')
        self.assertEqual(acl[1][1], 'system.Authenticated')
        self.assertEqual(sorted(acl[1][2]), ['view'])

        self.assertEqual(acl[2][0], 'Allow')
        self.assertEqual(acl[2][1], 'role:viewer')
        self.assertEqual(sorted(acl[2][2]), ['list', 'view'])

        self.assertEqual(acl[3][0], 'Allow')
        self.assertEqual(acl[3][1], 'role:editor')
        self.assertEqual(
            sorted(acl[3][2]),
            ['add', 'change_order', 'edit', 'list', 'view']
        )

        self.assertEqual(acl[4][0], 'Allow')
        self.assertEqual(acl[4][1], 'role:admin')
        self.assertEqual(sorted(acl[4][2]), [
            'add', 'change_order', 'change_state', 'copy', 'cut', 'delete',
            'edit', 'list', 'manage_permissions', 'paste', 'view'
        ])

        self.assertEqual(acl[5][0], 'Allow')
        self.assertEqual(acl[5][1], 'role:manager')
        self.assertEqual(sorted(acl[5][2]), [
            'add', 'change_order', 'change_state', 'copy', 'cut', 'delete',
            'edit', 'list', 'manage', 'manage_permissions', 'paste', 'view'
        ])

        self.assertEqual(acl[6][0], 'Allow')
        self.assertEqual(acl[6][1], 'role:owner')
        self.assertEqual(sorted(acl[6][2]), [
            'add', 'change_order', 'change_state', 'copy', 'cut', 'delete',
            'edit', 'list', 'manage_permissions', 'paste', 'view'
        ])

        self.assertEqual(acl[7][0], 'Allow')
        self.assertEqual(acl[7][1], 'system.Everyone')
        self.assertEqual(sorted(acl[7][2]), ['login'])

        self.assertEqual(acl[8][0], 'Deny')
        self.assertEqual(acl[8][1], 'system.Everyone')
        self.assertEqual(acl[8][2], ALL_PERMISSIONS)

    def test_ZODBEntryPrincipalACL(self):
        self.layer.new_request()
        node = ZODBPrincipalACLEntry()
        self.assertTrue(isinstance(node.storage, ZODBPrincipalACLEntryNode))

        self.assertTrue(isinstance(node.principal_roles, OOBTNodeAttributes))
        self.assertEqual(node.principal_roles.name, 'principal_roles')
        self.assertTrue(node.storage.principal_roles is node.principal_roles)

        node.principal_roles['someuser'] = ['manager']
        acl = node.__acl__

        self.assertTrue(node.storage.__acl__ == node.__acl__)

        self.assertEqual(acl[0][0], 'Allow')
        self.assertEqual(acl[0][1], 'someuser')
        self.assertEqual(sorted(acl[0][2]), [
            'add', 'change_order', 'change_state', 'copy', 'cut', 'delete',
            'edit', 'list', 'manage', 'manage_permissions', 'paste', 'view'
        ])

        self.assertEqual(acl[1][0], 'Allow')
        self.assertEqual(acl[1][1], 'system.Authenticated')
        self.assertEqual(sorted(acl[1][2]), ['view'])

        self.assertEqual(acl[2][0], 'Allow')
        self.assertEqual(acl[2][1], 'role:viewer')
        self.assertEqual(sorted(acl[2][2]), ['list', 'view'])

        self.assertEqual(acl[3][0], 'Allow')
        self.assertEqual(acl[3][1], 'role:editor')
        self.assertEqual(
            sorted(acl[3][2]),
            ['add', 'change_order', 'edit', 'list', 'view']
        )

        self.assertEqual(acl[4][0], 'Allow')
        self.assertEqual(acl[4][1], 'role:admin')
        self.assertEqual(sorted(acl[4][2]), [
            'add', 'change_order', 'change_state', 'copy', 'cut', 'delete',
            'edit', 'list', 'manage_permissions', 'paste', 'view'
        ])

        self.assertEqual(acl[5][0], 'Allow')
        self.assertEqual(acl[5][1], 'role:manager')
        self.assertEqual(sorted(acl[5][2]), [
            'add', 'change_order', 'change_state', 'copy', 'cut', 'delete',
            'edit', 'list', 'manage', 'manage_permissions', 'paste', 'view'
        ])

        self.assertEqual(acl[6][0], 'Allow')
        self.assertEqual(acl[6][1], 'role:owner')
        self.assertEqual(sorted(acl[6][2]), [
            'add', 'change_order', 'change_state', 'copy', 'cut', 'delete',
            'edit', 'list', 'manage_permissions', 'paste', 'view'
        ])

        self.assertEqual(acl[7][0], 'Allow')
        self.assertEqual(acl[7][1], 'system.Everyone')
        self.assertEqual(sorted(acl[7][2]), ['login'])

        self.assertEqual(acl[8][0], 'Deny')
        self.assertEqual(acl[8][1], 'system.Everyone')
        self.assertEqual(acl[8][2], ALL_PERMISSIONS)



from node.ext.zodb import IZODBNode
from node.interfaces import IMappingStorage
from node.interfaces import IUUIDAware


class IZODBEntry(IMappingStorage):
    """Marker interface for ZODB entry. ``storage`` attribute contains a
    ``IZODBEntryNode`` implementing instance.
    """


class IZODBEntryNode(IZODBNode):
    """Marker interface for first level ZODB nodes.
    """


class ICatalogAware(IUUIDAware):
    """Marker interface if node is ZODB catalog aware.
    """



from cone.app.interfaces import IWorkflowState
from cone.app.model import BaseNode
from cone.app.model import Metadata
from cone.zodb import app_path
from cone.zodb import combined_title
from cone.zodb import create_default_catalog
from cone.zodb import create_default_metadata
from cone.zodb import get_state
from cone.zodb import get_title
from cone.zodb import get_type
from cone.zodb import get_uid
from cone.zodb import str_app_path
from cone.zodb import str_zodb_path
from cone.zodb import testing
from cone.zodb import zodb_path
from cone.zodb import ZODBEntry
from cone.zodb.indexing import FLOORDATETIME
from cone.zodb.indexing import force_dt
from cone.zodb.testing import ZODBDummyNode
from datetime import datetime
from node.behaviors import UUIDAware
from node.tests import NodeTestCase
from plumber import plumbing
from zope.interface import alsoProvides
import uuid


class TestIndexing(NodeTestCase):
    layer = testing.zodb_layer

    def test_indexing(self):
        # Helper functions for catalog indexing
        root = BaseNode(name='root')
        entry = root['myentry'] = ZODBEntry()
        foo = entry['foo'] = ZODBDummyNode()
        bar = entry['bar'] = ZODBDummyNode()
        bar.attrs['title'] = 'Bar title'

        # ``path``
        self.assertEqual(foo.path, ['root', 'myentry', 'foo'])

        # ``zodb_path``
        self.assertEqual(zodb_path(entry), ['myentry'])
        self.assertEqual(zodb_path(entry.storage), ['myentry'])
        self.assertEqual(zodb_path(bar), ['myentry', 'bar'])
        self.assertEqual(zodb_path(foo), ['myentry', 'foo'])

        # ``str_zodb_path``
        self.assertEqual(str_zodb_path(bar), '/myentry/bar')
        self.assertEqual(str_zodb_path(foo), '/myentry/foo')

        # ``app_path``
        self.assertEqual(app_path(foo), ['root', 'myentry', 'foo'])

        # ``str_app_path``
        self.assertEqual(str_app_path(foo), '/root/myentry/foo')

        # ``combined_title``
        self.assertEqual(combined_title(bar), 'myentry - Bar title')

        # ``force_dt``
        self.assertEqual(force_dt('foo'), datetime(1980, 1, 1, 0, 0))
        self.assertTrue(FLOORDATETIME == force_dt('foo'))

        self.assertEqual(
            force_dt(datetime(2011, 5, 1)),
            datetime(2011, 5, 1, 0, 0)
        )

        # ``get_uid``
        @plumbing(UUIDAware)
        class UUIDNode(BaseNode):
            pass

        self.assertEqual(get_uid(BaseNode(), 'default'), 'default')
        self.assertTrue(isinstance(get_uid(UUIDNode(), 'default'), uuid.UUID))

        # ``get_type``
        self.assertEqual(get_type(object(), 'default'), 'default')
        self.assertEqual(get_type(foo, 'default'), 'dummytype')

        # ``get_state``
        self.assertEqual(get_state(object(), 'default'), 'default')

        setattr(foo, 'state', 'state_1')
        alsoProvides(foo, IWorkflowState)
        self.assertEqual(get_state(foo, 'default'), 'state_1')

        # ``get_title`` First ``node.metadata``, then ``node.attrs`` are
        # searched for title. finally default takes effect
        self.assertEqual(get_title(foo, 'default'), 'foo')
        self.assertEqual(get_title(BaseNode(), 'default'), u'no_title')

        class NoTitleNode(BaseNode):

            @property
            def metadata(self):
                return Metadata()

        self.assertEqual(get_title(NoTitleNode(), 'default'), 'default')

        # ``create_default_catalog``
        self.assertEqual(
            sorted(create_default_catalog().keys()),
            ['app_path', 'path', 'state', 'title', 'type', 'uid']
        )

        # ``create_default_metadata``
        setattr(bar, 'state', 'some_wf_state')
        alsoProvides(bar, IWorkflowState)
        self.assertEqual(create_default_metadata(bar), {
            'app_path': ['root', 'myentry', 'bar'],
            'uid': bar.uuid,
            'title': 'Bar title',
            'state': 'some_wf_state',
            'combined_title': 'myentry - Bar title',
            'path': ['myentry', 'bar']
        })



from cone.zodb.catalog import CatalogAware
from cone.zodb.catalog import CatalogIndexer
from cone.zodb.catalog import CatalogProvidingEntry
from cone.zodb.catalog import CatalogProxy
from cone.zodb.common import ZODBEntryPrincipalACL
from cone.zodb.common import ZODBPrincipalACL
from cone.zodb.entry import zodb_entry_for
from cone.zodb.entry import ZODBEntry
from cone.zodb.entry import ZODBEntryNode
from cone.zodb.entry import ZODBEntryStorage
from cone.zodb.indexing import app_path
from cone.zodb.indexing import combined_title
from cone.zodb.indexing import create_default_catalog
from cone.zodb.indexing import create_default_metadata
from cone.zodb.indexing import force_dt
from cone.zodb.indexing import get_state
from cone.zodb.indexing import get_title
from cone.zodb.indexing import get_type
from cone.zodb.indexing import get_uid
from cone.zodb.indexing import str_app_path
from cone.zodb.indexing import str_zodb_path
from cone.zodb.indexing import zodb_path
from cone.zodb.interfaces import ICatalogAware
from cone.zodb.interfaces import IZODBEntryNode



import sys
import unittest


def test_suite():
    from cone.zodb.tests import test_common
    from cone.zodb.tests import test_entry
    from cone.zodb.tests import test_indexing
    from cone.zodb.tests import test_catalog

    suite = unittest.TestSuite()

    suite.addTest(unittest.findTestCases(test_common))
    suite.addTest(unittest.findTestCases(test_entry))
    suite.addTest(unittest.findTestCases(test_indexing))
    suite.addTest(unittest.findTestCases(test_catalog))

    return suite


def run_tests():
    from zope.testrunner.runner import Runner

    runner = Runner(found_suites=[test_suite()])
    runner.run()
    sys.exit(int(runner.failed))


if __name__ == '__main__':
    run_tests()



from cone.zodb.entry import zodb_entry_for
from cone.zodb.indexing import create_default_catalog
from cone.zodb.indexing import create_default_metadata
from cone.zodb.interfaces import ICatalogAware
from cone.zodb.interfaces import IZODBEntryNode
from node.behaviors import UUIDAware
from plumber import Behavior
from plumber import default
from plumber import plumb
from repoze.catalog.document import DocumentMap
from repoze.catalog.interfaces import ICatalog
from repoze.catalog.query import Eq
from zope.interface import implementer
import uuid


class CatalogProxy(object):

    def __init__(self, dbcontext, indexing_root, catalog_key,
                 doc_map_key, catalog_factory, metadata_factory):
        """
        dbcontext
            IZODBNode instance.
        indexing_root
            Indexing root. defaults to self.dbcontext.
        catalog_key
            DB key of catalog.
        doc_map_key
            DB key of doc_map.
        catalog_factory
            Factory callback for creating catalog instance.
        metadata_factory
            Factory callback for creating node metadata.
        """
        self.dbcontext = dbcontext
        self.indexing_root = indexing_root
        self.catalog_key = catalog_key
        self.doc_map_key = doc_map_key
        self.catalog_factory = catalog_factory
        self.metadata_factory = metadata_factory

    @property
    def entry(self):
        return zodb_entry_for(self.dbcontext)

    @property
    def catalog(self):
        entry = self.entry
        catalog = entry.db_root.get(self.catalog_key)
        if catalog and not ICatalog.providedBy(catalog):
            raise ValueError(u'ICatalog not provided by %s' % self.catalog_key)
        if not catalog:
            self.reset_catalog()
            catalog = entry.db_root[self.catalog_key]
        return catalog

    @property
    def doc_map(self):
        entry = self.entry
        doc_map = entry.db_root.get(self.doc_map_key)
        if doc_map and not isinstance(doc_map, DocumentMap):
            raise ValueError(
                u'%s not a DocumentMap instance' % self.catalog_key)
        if not doc_map:
            entry.db_root[self.doc_map_key] = DocumentMap()
            doc_map = entry.db_root[self.doc_map_key]
        return doc_map

    def rebuild_catalog(self):
        self.reset_catalog()
        self._index_count = 0
        for child in self.indexing_root.values():
            self.index_recursiv(child)
        ret = self._index_count
        del self._index_count
        return ret

    def reset_catalog(self):
        self.entry.db_root[self.catalog_key] = self.catalog_factory()

    def doc_metadata(self, uid):
        if not isinstance(uid, uuid.UUID):
            uid = uuid.UUID(uid)
        doc_map = self.doc_map
        doc_id = doc_map.docid_for_address(uid)
        return doc_map.get_metadata(doc_id)

    def index_doc(self, node):
        uid = node.uuid
        doc_map = self.doc_map
        catalog = self.catalog
        indexed = catalog.query(Eq('uid', uid))
        if indexed[0]:
            docid = doc_map.docid_for_address(uid)
            catalog.unindex_doc(docid)
            doc_map.remove_address(uid)
        docid = doc_map.add(uid)
        metadata = self.metadata_factory(node)
        doc_map.add_metadata(docid, metadata)
        catalog.index_doc(docid, node)

    def index_recursiv(self, node):
        if ICatalogAware.providedBy(node):
            self.index_doc(node)
            if hasattr(self, '_index_count'):
                self._index_count += 1
        for child in node.values():
            self.index_recursiv(child)

    def unindex_doc(self, node):
        uid = node.uuid
        doc_map = self.doc_map
        docid = doc_map.docid_for_address(uid)
        self.catalog.unindex_doc(docid)
        doc_map.remove_address(uid)

    def unindex_recursiv(self, node):
        if ICatalogAware.providedBy(node):
            self.unindex_doc(node)
        for child in node.values():
            self.unindex_recursiv(child)


class CatalogIndexer(object):

    def __init__(self, proxies):
        self.proxies = proxies

    def _proxy(self, func_name, node):
        for proxy in self.proxies.values():
            getattr(proxy, func_name)(node)

    def index_doc(self, node):
        self._proxy('index_doc', node)

    def index_recursiv(self, node):
        self._proxy('index_recursiv', node)

    def unindex_doc(self, node):
        self._proxy('unindex_doc', node)

    def unindex_recursiv(self, node):
        self._proxy('unindex_recursiv', node)


@implementer(ICatalogAware)
class CatalogAware(UUIDAware):
    include_entry = default(False)
    """Flag controls whether to index entry node in calatog.
    """

    @default
    @property
    def catalog_proxies(self):
        proxies = dict()
        proxies['default'] = CatalogProxy(
            self, zodb_entry_for(self), 'cone_catalog', 'cone_doc_map',
            create_default_catalog, create_default_metadata)
        return proxies

    @default
    @property
    def catalog_indexer(self):
        return CatalogIndexer(self.catalog_proxies)

    @plumb
    def __init__(_next, self, *args, **kwargs):
        _next(self, *args, **kwargs)
        if self.include_entry and IZODBEntryNode.providedBy(self):
            self.catalog_indexer.index_doc(self)

    @plumb
    def __setitem__(_next, self, key, value):
        _next(self, key, value)
        self.catalog_indexer.index_recursiv(self[key])

    @plumb
    def __delitem__(_next, self, key):
        self.catalog_indexer.unindex_recursiv(self[key])
        _next(self, key)

    @plumb
    def __call__(_next, self):
        _next(self)
        if IZODBEntryNode.providedBy(self) and not self.include_entry:
            return
        self.catalog_indexer.index_doc(self)


class CatalogProvidingEntry(Behavior):
    """Helper behavior for ZODB entries to provide ``catalog_proxies`` and
    ``catalog_indexer`` directly.
    """

    @default
    @property
    def catalog_proxies(self):
        return self.storage.catalog_proxies

    @default
    @property
    def catalog_indexer(self):
        return CatalogIndexer(self.storage.catalog_proxies)



from cone.app.interfaces import IWorkflowState
from cone.app.model import AppRoot
from cone.zodb.entry import ZODBEntry
from cone.zodb.entry import ZODBEntryNode
from node.interfaces import IUUIDAware
from repoze.catalog.catalog import Catalog
from repoze.catalog.indexes.field import CatalogFieldIndex
from repoze.catalog.indexes.path import CatalogPathIndex
import datetime


FLOORDATETIME = datetime.datetime(1980, 1, 1)  # XXX tzinfo


def force_dt(value):
    if not isinstance(value, datetime.datetime):
        return FLOORDATETIME
    return value


def zodb_path(node, default=None):
    path = list()
    while True:
        if isinstance(node, ZODBEntryNode):
            node = node.entry
        path.append(node.name)
        if node.parent is None or isinstance(node, ZODBEntry):
            path.reverse()
            return path
        node = node.parent


def str_zodb_path(node, default=None):
    path = zodb_path(node, default)
    if path:
        return '/%s' % '/'.join(path)


def app_path(node, default=None):
    path = list()
    while True:
        path.append(node.name)
        if node.parent is None or isinstance(node.parent, AppRoot):
            path.reverse()
            return path
        node = node.parent


def str_app_path(node, default=None):
    path = app_path(node, default)
    if path:
        return '/%s' % '/'.join(path)


def get_uid(node, default):
    return IUUIDAware.providedBy(node) and node.uuid or default


def get_type(node, default):
    if hasattr(node, 'node_info_name'):
        return node.node_info_name
    return default


def get_state(node, default):
    if hasattr(node, 'state'):
        return node.state
    return default


def get_title(node, default):
    if hasattr(node, 'metadata'):
        title = node.metadata.title
        if title:
            return title
    return node.attrs.get('title', default)


def combined_title(node):
    titles = list()
    while True:
        titles.append(node.attrs.get('title', node.name))
        if node.parent is None or isinstance(node, (ZODBEntry, ZODBEntryNode)):
            break
        node = node.parent
    titles.reverse()
    return ' - '.join(titles)


def create_default_catalog():
    catalog = Catalog()
    catalog['uid'] = CatalogFieldIndex(get_uid)
    catalog['type'] = CatalogFieldIndex(get_type)
    catalog['state'] = CatalogFieldIndex(get_state)
    catalog['path'] = CatalogPathIndex(str_zodb_path)
    catalog['app_path'] = CatalogPathIndex(str_app_path)
    catalog['title'] = CatalogFieldIndex(get_title)
    return catalog


def create_default_metadata(node):
    metadata = dict()
    metadata['uid'] = node.uuid
    metadata['path'] = zodb_path(node)
    metadata['app_path'] = app_path(node)
    metadata['title'] = get_title(node, node.name)
    metadata['combined_title'] = combined_title(node)
    if IWorkflowState.providedBy(node):
        metadata['state'] = node.state
    return metadata
